// #define TESTER_CUSTOM //     
// #define TESTER_CUSTOM_SLAVE //       "" 

#include <fxsaber\Expert.mqh> // https://www.mql5.com/ru/code/19003

#define __TESTER__

#include "TesterBase.mqh"
#include "Order_HLines.mqh"
#include "History_Line.mqh"

#include <Report.mqh> // https://www.mql5.com/ru/code/18801

#define SPEED_KOEF 1.5

class TESTER : public TESTERBASE
{
private:
  HISTORY_LINE HistoryLine;

  void UpdateLines( const bool FullHistory = false )
  {
    if (FullHistory)
      this.HistoryLine.Clear();
    
    ORDER_HLINES HLines;

    HLines.Sync();
    this.HistoryLine.Check();

    return;
  }

  static bool HistoryIsReady( const datetime tBeginTime, MqlTick &BeginTick )
  {
    MqlTick Ticks[1];
  
    ::ResetLastError();
  
    if (::CopyTicks(_Symbol, Ticks, COPY_TICKS_ALL, (ulong)tBeginTime * 1000, 1) == 1)
      BeginTick = Ticks[0];
  
    return(!::GetLastError() ||
           (!::TerminalInfoInteger(TERMINAL_CONNECTED) && 
           ((::CopyTicks(_Symbol, Ticks, COPY_TICKS_ALL, (ulong)(tBeginTime - 7 * 24 * 3600) * 1000, 1)) == 1) &&
           (Ticks[0].time < tBeginTime)));
  }

  void OnTimerReal()
  {
    static bool FirstRun = true;

    MqlTick BeginTick = {0};

    if (!TESTER::HistoryIsReady(this.StartTime, BeginTick))
    {            
      const datetime LastTime2 = (datetime)::SymbolInfoInteger(_Symbol, SYMBOL_TIME);
      const datetime EndTime = ::TimeCurrent() > LastTime2 ? LastTime2 : ::TimeCurrent();

      ::Comment("History = " + DoubleToString(100.0 * (EndTime - BeginTick.time) / (EndTime - this.StartTime), 2) + "%");
    }
    else if (FirstRun)
    {
      FirstRun = false; //   ,    .
      ::EventKillTimer();

      BeginTick.time_msc = (long)StartTime * 1000;
      BeginTick.time = StartTime;

      this.Symb.DeleteHistory();
      this.Symb.Delete(true);
      this.Symb.Create("Tester", _Symbol);

      if (this.Symb.IsExist()) //   
      {        
        this.Symb.On();
        
        //  10 M1-  
        MqlRates Rates[];        
        ::CopyRates(_Symbol, PERIOD_M1, this.StartTime, 10000, Rates);
        this.Symb += Rates;

        string Names[];
        MqlParam Params[];

        if ((this.Symb += BeginTick) && EXPERT::Parameters(0, Params, Names))
        {          
          Params[0].string_value += "\nexpertmode=5"; //    DLL
  
          for (int i = ::ArraySize(Names) - 1; i >= 0; i--)
          {
            if (Names[i] == "inStartTime")
              Params[i + 1].string_value = (string)(-::ChartID());
  
            EXPERT::AddInputName(Params[i + 1], Names[i]);
          }
  
          EXPERT::Run(::ChartOpen(this.Symb.Name, _Period), Params);
  
          ::Comment("");
  
          //  -       
          if (::MQLInfoInteger(MQL_DEBUG))
            ::Sleep(10000);
        }

        ::ExpertRemove();
      }
    }

    return;
  }


  static bool IsKeyEsc( void )
  {
    return(::TerminalInfoInteger(TERMINAL_KEYSTATE_ESCAPE) < 0);
  }

  static bool IsKeyTab( void )
  {
    return(::TerminalInfoInteger(TERMINAL_KEYSTATE_TAB) < 0);
  }

  static bool IsKeyCtrl( void )
  {
    return(::TerminalInfoInteger(TERMINAL_KEYSTATE_CONTROL) < 0);
  }

  static bool IsKeyShift( void )
  {
    return(::TerminalInfoInteger(TERMINAL_KEYSTATE_SHIFT) < 0);
  }
/*
#define TERMINAL_KEYSTATE_ALT 1018
  static bool IsKeyAlt( void )
  {
    return(::TerminalInfoInteger(TERMINAL_KEYSTATE_ALT) < 0);
  }
#undef TERMINAL_KEYSTATE_ALT
*/
  void NewTickOffset( const int BidOffset, const int AskOffset )
  {
    MqlTick Tick[1];
    
    if (::SymbolInfoTick(this.Symb.Name, Tick[0]) && Tick[0].bid && Tick[0].ask)
    {
      const double point = ::SymbolInfoDouble(this.Symb.Name, SYMBOL_POINT);
      const int digits = (int)::SymbolInfoInteger(this.Symb.Name, SYMBOL_DIGITS);
      
      Tick[0].bid += BidOffset * point;
      Tick[0].ask += AskOffset * point;
      
/*      
      // https://www.mql5.com/ru/forum/304554/page13#comment_12034266
      ::NormalizeDouble(Tick[0].bid, digits);
      ::NormalizeDouble(Tick[0].ask, digits);
*/
      
      this.NewTicks(Tick);
    }
    
    return;
  }
  

#ifdef TESTER_CUSTOM_SLAVE  
  virtual void NewTicks( const int, const int Interval )
  {
    static MqlTick NewTicks[1];
    
    if (::SymbolInfoTick(_Symbol, NewTicks[0]))
      this.System(NewTicks);

    return;
  }
#endif // TESTER_CUSTOM_SLAVE
  
public:
  const bool IsTester;

#ifdef TESTER_CUSTOM_SLAVE
  TESTER( const datetime dStartTime, const STRATEGY sStrategy = NULL ) : TESTERBASE(_Symbol,
                                                                                   (datetime)::SymbolInfoInteger(_Symbol, SYMBOL_TIME),
                                                                                   NULL,
                                                                                   sStrategy),
                                                                         IsTester(true)
#else // TESTER_CUSTOM_SLAVE
  TESTER( const datetime dStartTime, const STRATEGY sStrategy = NULL ) : TESTERBASE(((dStartTime < 0) ? NULL : "TESTER_") + _Symbol,                                                                                   
                                                                                   (dStartTime < 0) ? (datetime)_V(0, ::SymbolInfoInteger(_Symbol, SYMBOL_TIME)) : dStartTime,
                                                                                   (dStartTime < 0) ? ::ChartSymbol(-dStartTime) : _Symbol,
                                                                                   sStrategy),
                                                                         IsTester(dStartTime < 0)
#endif // TESTER_CUSTOM_SLAVE
  {
    if (this.IsTester)
    {
      ::ChartSetInteger(0, CHART_SHOW_OBJECT_DESCR, true);

      ::Print("EventSetMillisecondTimer - Waiting.");
      
      if (::EventSetMillisecondTimer(15))
        ::Print("EventSetMillisecondTimer - Done.");
      else
      {
        ::Alert("EventSetMillisecondTimer - ERROR!");
        
        ::ExpertRemove();
      }

      this.Symb.DeleteTicks(this.StartTime * 1000); //   ,   SymbolInfoInteger(_Symbol, SYMBOL_TIME)  TESTERBASE.        
      
      this.NextVirtual(false);
    }
    else
    {
      ::Print("EventSetTimer - Waiting.");
      
      if (::EventSetTimer(1))
        ::Print("EventSetTimer - Done.");
      else
      {
        ::Alert("EventSetMillisecondTimer - ERROR!");
        
        ::ExpertRemove();
      }
    }

    this.OnTimer();    
  }

  ~TESTER( void )
  {
    ::EventKillTimer();
  }

  bool OnTimer( void )
  {
    if (!::IsStopped())
    {
      if (this.IsTester)
      {
        this.OnTimerTester();

        this.UpdateLines(); //       ,   .
        ::ChartRedraw();
      }
      else
      {
        ::Comment("Waiting history....");

        this.OnTimerReal();
      }
    }

    return(true);
  }

#define KEY_BREAK              19
#define KEY_UP                 38
#define KEY_DOWN               40
#define KEY_F5                116
#define KEY_R                  82
#define KEY_LEFT               37

  void OnChartEvent( const int id, const long &lparam, const double &dparam, const string &sparam )
  {
    if (this.IsTester && (id == CHARTEVENT_KEYDOWN))
    {
      switch ((int)lparam)
      {
      case KEY_BREAK: // 
        this.ChangePauseMode();

        break;
      case KEY_UP:
        if (TESTER::IsKeyEsc())
          ::TesterDeposit(1000); // UP+ESC - Deposit
        else if (TESTER::IsKeyTab() && this.GetPauseMode())
          this.NewTickOffset(1, 1); // UP+TAB - Bid/Ask += Point
        else if (TESTER::IsKeyShift() && this.GetPauseMode())
          this.NewTickOffset(0, 1); // UP+SHIFT - Ask += Point
        else if (TESTER::IsKeyCtrl() && this.GetPauseMode())
          this.NewTickOffset(1, 0); // UP+CTRL - Bid += Point
        else
          this.ChangeVelocity(SPEED_KOEF); //  

        break;
      case KEY_DOWN:
        if (TESTER::IsKeyEsc())
          ::TesterWithdrawal(1000); // // DOWN+ESC - Withdrawal
        else if (TESTER::IsKeyTab() && this.GetPauseMode())
          this.NewTickOffset(-1, -1); // DOWN+TAB - Bid/Ask -= Point
        else if (TESTER::IsKeyShift() && this.GetPauseMode())
          this.NewTickOffset(0, -1); // DOWN+SHIFT - Bid -= Point
        else if (TESTER::IsKeyCtrl() && this.GetPauseMode())
          this.NewTickOffset(-1, 0); // DOWN+CTRL - Bid -= Point
        else
          this.ChangeVelocity(1 / SPEED_KOEF); //  
        
//        ::Sleep(100); //   .                  
        ::ChartSetSymbolPeriod(0, _Symbol, _Period); //    .       
        ::ChartRedraw();

        break;
      case KEY_F5: //  
        if (this.GetPauseMode())
        {
          this.NewTicks(1);

          this.UpdateLines();
          ::ChartRedraw();
        }

        break;
      case KEY_R: //  
        if (this.GetPauseMode())
        {
          REPORT::ToFile("Tester\\" + ::MQLInfoString(MQL_PROGRAM_NAME) + "\\Report.htm");          
          ::ChartRedraw();
        }

        break;
      case KEY_LEFT: //
        if (this.GetPauseMode() && TESTER::IsKeyCtrl())
        {
          this.PrevVirtual();          
          
          this.UpdateLines(true);

//          ::Sleep(100); //   .          
          ::ChartSetSymbolPeriod(0, _Symbol, _Period); //    .
          ::ChartRedraw();
        }

        break;
      }
    }
  }

#undef KEY_LEFT
#undef KEY_R
#undef KEY_F5
#undef KEY_DOWN
#undef KEY_UP
#undef KEY_BREAK
};

#undef SPEED_KOEF

//************************
#ifdef TESTER_CUSTOM

sinput datetime inStartTime = D'2019.06.01 00:00';

void OnTick()
{
  static const bool IsTester = (inStartTime < 0);

  if (IsTester)
    OnTick_Tester();
}

TESTER CustomTester(inStartTime, ::OnTick); // https://www.mql5.com/ru/forum/304239/page15#comment_10781426

void OnTimer()
{
  CustomTester.OnTimer();

  if (CustomTester.IsTester)
    ::OnTimer_Tester();
}
#define OnTimer OnTimer_Tester

void OnChartEvent( const int id, const long &lparam, const double &dparam, const string &sparam )
{
  CustomTester.OnChartEvent(id, lparam, dparam, sparam);

  if (CustomTester.IsTester)
    ::OnChartEvent_Tester(id, lparam, dparam, sparam);
}
#define OnChartEvent OnChartEvent_Tester

int OnInit( void )
{
  return(CustomTester.IsTester ? OnInit_Tester() : INIT_SUCCEEDED);
}
#define OnInit OnInit_Tester

void OnDeinit( const int Reason )
{
  if (CustomTester.IsTester)
    ::OnDeinit_Tester(Reason);
}
#define OnDeinit OnDeinit_Tester

bool EventSetTimer_Tester( const int ) { return(true); }
void EventKillTimer_Tester( void ) { return; }
#define EventKillTimer EventKillTimer_Tester
#define EventSetTimer EventSetTimer_Tester
#define EventSetMillisecondTimer EventSetTimer_Tester

#define OnTrade OnTrade_Tester
#define OnTradeTransaction OnTradeTransaction_Tester
#define OnTick OnTick_Tester

/*
void OnChartEvent( const int id, const long &lparam, const double &dparam, const string &sparam ) {}
void OnTick() {}
void OnTimer() {}
int OnInit() { return(INIT_SUCCEEDED); }
void OnDeinit( const int ) {}
*/
#endif // TESTER_CUSTOM